#!/bin/ksh -r
#
#	$Id: cagair.ksh,v 1.46 2014/11/24 21:33:28 craski-shell_86 Exp $
#
# vim: tabstop=4 shiftwidth=4 softtabstop=4 noexpandtab
#
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
#
# Copyright (c) 2008 Christian Weisgerber. Public Domain.
#
# Copyright (c) 2014 Craig Skinner <skinner@britvault.co.uk>
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
#


. ${TRUEPREFIX}/lib/sysexits.sh


THIS='${PROJECT}'
REPO="${HOME}/.${THIS}"
THAT="${REPO}/.${THIS}!${LOGNAME}"
CONF="${THAT}.rc"
SAFE="${THAT}.txt.dat"
LOCK="${TMPDIR:-/tmp}/${LOGNAME}.${THIS}.lock"
PIDF="${LOCK}/${THIS}.pid"
VRSN='${VERSION}'
EDITORS[0]='mg'
EDITORS[1]='vi'
EDITORS[2]='vim'

readonly THIS REPO THAT CONF SAFE LOCK PIDF VRSN EDITORS


function bail_out
{
	[[ -d ${TMPDIR} ]] && rm -RP ${TMPDIR}
	rm -rf ${LOCK}
	[[ ${1} -eq 0 ]] && clear
	exit ${1}
}


function make_temp
{
	for dir in ${TMPDIR} /tmp /var/tmp ${HOME}
	do
		export TMPDIR=${dir}
		_temp=$(mktemp ${1}) && break
		unset _temp
	done

	[[ -f ${_temp} || -d ${_temp} ]] &&
	{
		touch ${_temp} &&
		{
			chflags nodump ${_temp}
			[[ -f ${_temp} ]] && export TMPFILE=${_temp}
			[[ -d ${_temp} ]] && export TMPDIR=${_temp}
			return
		}
	}

	print -u2 'Unable to create a temporay file/directory'
	bail_out ${EX_CANTCREAT}
}


function set_mode
{
	local _mode=${1}
	[[ -n ${_mode} && $# -gt 1 ]] ||
	{
		print -u2 "usage: $0() 6-digit-st_mode file/dir ..."
		bail_out ${EX_SOFTWARE}
	}

	shift
	for _item in $*
	do
		[[ -f ${_item} || -d ${_item} ]] || continue

		export $(stat -s ${_item})
		[[ ${st_mode} -eq ${_mode} ]] ||
		{
			chmod ${_mode} ${_item} || bail_out ${EX_NOPERM}
		}
	done
}


function dencrypt_file
{
	[[ $# -eq 3 ]] || return ${EX_SOFTWARE}
	[[ ${1} == '-d' || ${1} == '-e' ]] || return ${EX_SOFTWARE}
	[[ -f ${2} && -f ${3} ]] || return ${EX_IOERR}

	print -n -- "${PASSWORD}" | \
		openssl aes-256-cbc ${1} -in ${2} -out ${3} \
				-pass stdin || return ${EX_DATAERR}
}


function get_editor
{
	# Remove any of the user's editor flags:
	unalias ${EDVIS} ${EDITOR} ${VISUAL}

	[[ -f ${CONF} && -s ${CONF} ]] && . ${CONF}

	EDVIS=${EDVIS-${EDITOR-${VISUAL}}}
	[[ -n ${EDVIS} ]] &&
	{
		for _editor in ${EDITORS[*]}
		do
			[[ ${EDVIS##*/} == ${_editor} ]] &&
			{
				whence -p ${_editor} &&
				{
					EDVIS=${_editor}
					read choice?'Change editor? [y/N]: '
					[[ ${choice} == @(y|Y) ]] || return
					break
				}
			}
		done
		unset EDVIS
	}

	print 'Secured file editors are:'
	for _editor in ${EDITORS[*]}
	do
		print -n "${_editor}:\t"
		whence -p ${_editor}
	done

	until [[ -n ${EDVIS} ]]
	do
		read choice?"Select from above list? [Default is ${EDITORS[0]}]: "
		for _editor in ${EDITORS[*]}
		do
			[[ ${choice:-${EDITORS[0]}} == ${_editor} ]] &&
			{
				whence -p ${_editor} &&
				{
					EDVIS=${_editor}
					break
				}
			}
		done
	done

	set_editor
}


function set_editor
{
	read choice?'Save selected editor for next time? [Y/n]: '
	[[ ${choice} == @(n|N) ]] ||
	{
		[[ -f ${CONF} ]] || touch ${CONF}

		grep -q ^EDVIS= ${CONF} ||
		{
			print "H\na\nEDVIS='${EDVIS}'\n.\nw" | ed -s ${CONF}
			return
		}

		print "H\n,s/^EDVIS=.*/EDVIS='${EDVIS}'/\n.\nw" | ed -s ${CONF}
	}
}


function secure_editor
{
	case ${EDVIS} in
		mg)
			# By default, mg sets 'make-backup-files' and
			# 'leave-tmpdir-backups', in ${TMPDIR} - OK
			# Users can set 'backup-to-home-directory' in ~/.mg
			# "-n	Turn off backup file generation."
			readonly EDVIS_ARGS='-n -f column-number-mode'
			;;
		vi)
			# Set vi's clear text file recovery directory to ${TMPDIR}
			readonly NEXINIT="set secure recdir=${TMPDIR}"
			export NEXINIT
			;;
		vim)
			# Store vim's temporary files in ${TMPDIR}
			VIMINIT="set secure directory=${TMPDIR}//"
			VIMINIT="${VIMINIT} backupdir=${TMPDIR}//"
			VIMINIT="${VIMINIT} undodir=${TMPDIR}//"
			VIMINIT="${VIMINIT} viminfo+=n${TMPDIR}//viminfo"
			readonly VIMINIT
			export VIMINIT
			# -i {viminfo} ... used to skip the use of the .viminfo file
			#
			# -n	No swap file will be used. Recovery
			#	after a crash will be impossible.
			#
			# -Z	Restricted mode.
			readonly EDVIS_ARGS='-n -i NONE -Z'
			;;
		*)
			print -u2 "Unable to secure editor '${EDVIS}'"
			bail_out ${EX_CONFIG}
			;;
	esac
}


function get_password
{
	STTY=$(stty -g)
	print -n 'Password: '
	stty -echo
	read PASSWORD
	stty "$STTY"
}


function check_password
{
	[[ -n ${PASSWORD_CHECKER} ]] &&
	{
		echo ${PASSWORD} | ${PASSWORD_CHECKER} || unset PASSWORD
	}

	[[ -n ${PASSWORD} ]] &&
	{
		PASSWORD_1=${PASSWORD}
		get_password
		[[ ${PASSWORD} == ${PASSWORD_1} ]] ||
		{
			print -u2 'They do not match.'
			unset PASSWORD PASSWORD_1
		}
	}
}


function usage
{
	_usage="$0:\n\t${THIS}\n"
	_usage="${_usage}\t${THIS} [-hv] [-i file]\n"
	_usage="${_usage}\t${THIS} -h (this help)\n"
	_usage="${_usage}\t${THIS} -i file (plain text import)\n"
	_usage="${_usage}\t${THIS} -v (version number)\n"
	_usage="${_usage}\nFor more help and Scottish Gaelic pronunciation,\n"
	_usage="${_usage}refer to the included README and MP3 files in:\n"
	_usage="${_usage}${TRUEPREFIX}/${DOC_DIR}/"
	[[ ${1} -gt 0 ]] && _print_args='-u2'
	print ${_print_args} -- ${_usage}
	exit ${1}
}


function check_import_file
{
	[[ -f ${1} ]] ||
	{
		print -u2 "Not a file to import: ${1}"
		return ${EX_USAGE}
	}
	[[ -r ${1} ]] ||
	{
		print -u2 "Can't read import file: ${1}"
		return ${EX_NOPERM}
	}
	[[ -s ${1} ]] ||
	{
		print -u2 "Empty import file: ${1}"
		return ${EX_DATAERR}
	}
	file ${1} | fgrep -q -w text ||
	{
		print -u2 "Not a text file to import: ${1}"
		return ${EX_DATAERR}
	}
}


# Alter/check the shell
trap 'bail_out' HUP INT QUIT ABRT TERM STOP
umask 077
for builtin in 'print' '[[' '[' 'whence'
do
	type ${builtin} | egrep -q ' is a shell builtin$| is a reserved word$' ||
	{
		print -u2 '*** Security breach! ***'
		print -u2 "The '${builtin}' command is not a shell builtin."
		bail_out ${EX_UNAVAILABLE}
	}
done


while getopts "hi:v" opt
do
	case ${opt} in
		h)
			usage
			;;
		i)
			import_file=${OPTARG}
			check_import_file ${import_file} || bail_out $?
			;;
		v)
			print "Version: ${VRSN}"
			exit
			;;
		\?)
			usage ${EX_USAGE}
			;;
	esac
done


mkdir ${LOCK} ||
{
	[[ -f ${PIDF} ]] &&
	{
		_pid=$(tail -n 1 ${PIDF})
		[[ -n ${_pid} ]] &&
		{
			ps -o user,command -p ${_pid} |
				egrep -q "^${LOGNAME} $(head -n 1 $0 | cut -d! -f2) $0" &&
				{
					print -u2 "${THIS} already running (PID: ${_pid})"
					exit ${EX_TEMPFAIL}
				}
		}
	}
}
touch ${PIDF} || bail_out $?
print "H\na\n$$\n.\nw" | ed -s ${PIDF} || bail_out $?


[[ -d ${REPO} ]] ||
{
	mkdir ${REPO} || bail_out ${EX_CANTCREAT}
}
set_mode 040300 ${REPO}
set_mode 100600 ${SAFE} ${CONF}
make_temp -d
make_temp
get_editor
secure_editor


LOGIN_CLASS=$(userinfo ${LOGNAME} | awk '/^class/ { print $2 }')
PASSWORD_CHECKER=$(getcap -f ${SYSCONFDIR}/login.conf -c passwordcheck ${LOGIN_CLASS:-default})
[[ -n ${PASSWORD_CHECKER} ]] &&
{
	PASSWORD_CHECKER=$(print ${PASSWORD_CHECKER} | cut -d= -f2)
	# Restricted shell; ensure it is in our $PATH
	PASSWORD_CHECKER=${PASSWORD_CHECKER##*/}
	whence -p $(print ${PASSWORD_CHECKER} | awk '{ print $1 }') || unset PASSWORD_CHECKER
}


trap 'bail_out $?' ERR


if [[ -f ${SAFE} ]]
then
	print 'Decrypting existing secrets...'
	get_password
	dencrypt_file -d ${SAFE} ${TMPFILE}
else
	touch ${SAFE}
fi
TMPFILE_CKSUM_BEFORE_EDIT=$(cksum ${TMPFILE})

[[ -n ${import_file} ]] &&
{
	print "H\na\n\n# import of ${import_file} @ $(date)\n\n.\nw" | ed -s ${TMPFILE}
	cp -p ${TMPFILE} ${TMPFILE}~
	sed -n w${TMPFILE} ${TMPFILE}~ ${import_file}
	print -n '\nImport file was appended. Continue [ENTER]: '
	read
}

${EDVIS} ${EDVIS_ARGS} ${TMPFILE}
[[ ! -s ${SAFE} && ! -s ${TMPFILE} ]] &&
{
	print 'Nothing to encrypt, exiting.'
	rm ${SAFE} ${TMPFILE}
	bail_out ${EX_OK}
}

TMPFILE_CKSUM_AFTER_EDIT=$(cksum ${TMPFILE})
[[ ${TMPFILE_CKSUM_BEFORE_EDIT} == ${TMPFILE_CKSUM_AFTER_EDIT} ]] &&
{
	print 'No modification, exiting.'
	bail_out ${EX_OK}
}

print 'Encrypting secrets...'
[[ -n ${PASSWORD} ]] &&
{
	read choice?'Reuse the same password again? [Y/n]: '
	[[ ${choice} == @(n|N) ]] && unset PASSWORD
}
[[ -z ${PASSWORD} ]] &&
{
	WARNING='\n*** WARNING!!! ***\n'
	WARNING="${WARNING}There is _ABSOLUTELY_NO_ password recovery procedure!\n"
	WARNING="${WARNING}*** Be VERY careful to REMEMBER this! ***\n"
	print ${WARNING}
	while [[ -z ${PASSWORD} ]]
	do
		get_password
		check_password
	done
}
dencrypt_file -e ${TMPFILE} ${SAFE}
bail_out

